import { ExprValueTuples, DataTypeInputs, Expr, ExprValueObjects } from '@typed-query-builder/builder';
import { DialectOutput } from '@typed-query-builder/sql';
import { ConnectionPool, IResult, ISqlType, PreparedStatement, Request, Transaction } from 'mssql';
import './parsers';
/**
 * Options
 */
export interface MssqlOptions<P> {
    /**
     * Named parameters to pass to the query.
     */
    params?: P;
    /**
     * If errors should be thrown when building the SQL when an unsupported feature is being used.
     */
    throwError?: boolean;
    /**
     * If constants should avoided being passed in SQL string and passed as parameters instead.
     */
    constantsAsParams?: boolean;
    /**
     * If the raw expression should be executed, otherwise if it's not a SELECT, INSERT, DELETE, or UPDATE it will be wrapped as a SELECT.
     */
    raw?: boolean;
    /**
     * If the SQL string generated should remove redudant table/source references to columns that are unique in their context.
     */
    simplifyReferences?: boolean;
    /**
     * If the affected count should be returned, transforming the result into { affected: number, result: any }.
     */
    affectedCount?: boolean;
    /**
     * If any records to be returned should be arrays instead of objects.
     */
    arrayMode?: boolean;
    /**
     * When calling stream, if the stream should take pauses between batch sizes to process a batch at once.
     */
    streamBatchSize?: number;
    /**
     * If true all strings in results will be inspected for JSON values and be automatically parsed.
     */
    detectJson?: boolean;
    /**
     * If true all strings in results will be inspected for "YYYY-MM-DD" format and automatically converted to a Date (or custom object).
     */
    detectDate?: boolean | ((detected: string) => any);
    /**
     * If true all strings in results will be inspected for "YYYY-MM-DD(T| )hh:mm:ss" format and automatically converted to a Date (or custom object).
     */
    detectSmallDateTime?: boolean | ((detected: string) => any);
    /**
     * If true all strings in results will be inspected for "YYYY-MM-DD(T| )hh:mm:ss[.nnn]" format and automatically converted to a Date (or custom object).
     */
    detectDateTime?: boolean | ((detected: string) => any);
    /**
     * If true all strings in results will be inspected for "YYYY-MM-DD(T| )hh:mm:ss[.nnnnnnn]" format and automatically converted to a Date (or custom object).
     */
    detectDateTime2?: boolean | ((detected: string) => any);
    /**
     * If true all strings in results will be inspected for "YYYY-MM-DD(T| )hh:mm:ss[.nnnnnnn] [+|-]hh:mm" format and automatically converted to a Date (or custom object).
     */
    detectDateTimeOffset?: boolean | ((detected: string) => any);
    /**
     * If true all strings will be inspected for a date format and will be automatically converted to a Date (or custom object).
     */
    detectAllDates?: boolean | ((detected: string) => any);
}
export declare type AffectedResult<R> = {
    affected: number;
    result: R;
};
/**
 * Executes an expression and returns the result.
 *
 * **Example:**
 * ```ts
 * const result = await expr.run( exec(conn) );
 * ```
 *
 * @param access The connection, transaction, or prepared statement to stream the expression results from.
 * @param options Options that control how the query is built or the results returned.
 */
export declare function exec<P = any>(access: ConnectionPool | Transaction | PreparedStatement | undefined, options: MssqlOptions<P> & {
    affectedCount: true;
}): <R>(expr: Expr<R>) => Promise<AffectedResult<ExprValueObjects<R>>>;
export declare function exec<P = any>(access: ConnectionPool | Transaction | PreparedStatement | undefined, options: MssqlOptions<P> & {
    affectedCount: true;
    arrayMode: true;
}): <R>(expr: Expr<R>) => Promise<AffectedResult<ExprValueTuples<R>>>;
export declare function exec<P = any>(access: ConnectionPool | Transaction | PreparedStatement | undefined, options: MssqlOptions<P> & {
    arrayMode: true;
}): <R>(expr: Expr<R>) => Promise<ExprValueTuples<R>>;
export declare function exec<P = any>(access?: ConnectionPool | Transaction | PreparedStatement, options?: MssqlOptions<P>): <R>(expr: Expr<R>) => Promise<ExprValueObjects<R>>;
/**
 * The result passed to a stream listener.
 */
export declare type StreamListenerResult<R> = R extends Array<infer E> ? E : R;
/**
 * A listener to results.
 */
export declare type StreamListener<R, A> = (result: StreamListenerResult<R>, batchIndex: number, batchCount: number, batch: StreamListenerResult<R>[]) => A | Promise<A>;
/**
 * A handler that triggers the query and each result is passed to the listener.
 */
export declare type StreamHandler<R> = <A>(listener: StreamListener<R, A>) => Promise<A[]>;
/**
 * Generates a way to stream the results of a query for the processing of large datasets.
 * When batch size is given > 1, the given number are collected in memory and once that size is
 * reached or there are no more results the request is paused to avoid memory exhaustion and calls
 * the function in quick succession with all collected results.
 *
 * **Example:**
 * ```ts
 * const streamer = expr.run( stream(conn, { streamBatchSize: 100 }) );
 *
 * const accumulated = await streamer(async (record) => {
 *  // handle record, can be async, can return a value to be returned in accumulated array
 * });
 * ```
 *
 * @param access The connection, transaction, or prepared statement to stream the expression results from.
 * @param options Options that control how the query is built or the results returned.
 * @returns A function which when invoked with another function will execute the expression and for each result returned will call the given function.
 */
export declare function stream<P = any>(access: ConnectionPool | Transaction | PreparedStatement | undefined, options: MssqlOptions<P> & {
    arrayMode: true;
}): <R>(expr: Expr<R>) => StreamHandler<ExprValueTuples<R>>;
export declare function stream<P = any>(access?: ConnectionPool | Transaction | PreparedStatement, options?: MssqlOptions<P>): <R>(expr: Expr<R>) => StreamHandler<ExprValueObjects<R>>;
/**
 * A prepared query that can be executed multiple times. It MUST be released, ideally in a try-finally block.
 */
export interface PreparedQuery<R, P = any> {
    exec(params: P): Promise<R>;
    release(): Promise<void>;
}
/**
 * Creates a prepared statement for the given expression. This is useful when you need to invoke the same query over and over
 * with different parameters. A prepared statement MUST be released, ideally in a try-finally block.
 *
 * **Example:**
 * ```ts
 * const prepared = expr.run( prepare(conn) );
 *
 * try {
 *  await prepared.exec({ id: 12 });
 *  await prepared.exec({ id: 34 });
 * } finally {
 *  await prepared.release();
 * }
 * ```
 *
 * @param access The connection, transaction, or prepared statement to stream the expression results from.
 * @param options Options that control how the query is built or the results returned.
 * @returns An object that can be executed multiple times, once finished it must be released.
 */
export declare function prepare<P = any>(access: ConnectionPool | Transaction | undefined, options: MssqlOptions<P> & {
    affectedCount: true;
}): <R>(expr: Expr<any>) => Promise<PreparedQuery<AffectedResult<ExprValueObjects<R>>, P>>;
export declare function prepare<P = any>(access: ConnectionPool | Transaction | undefined, options: MssqlOptions<P> & {
    affectedCount: true;
    arrayMode: true;
}): <R>(expr: Expr<any>) => Promise<PreparedQuery<AffectedResult<ExprValueTuples<R>>, P>>;
export declare function prepare<P = any>(access: ConnectionPool | Transaction | undefined, options: MssqlOptions<P> & {
    arrayMode: true;
}): <R>(expr: Expr<R>) => Promise<PreparedQuery<ExprValueTuples<R>, P>>;
export declare function prepare<P = any>(access?: ConnectionPool | Transaction, options?: MssqlOptions<P>): <R>(expr: Expr<R>) => Promise<PreparedQuery<ExprValueObjects<R>, P>>;
export declare function parseResult<R, P>(expr: Expr<R>, iresult: IResult<ExprValueObjects<R>>, options?: MssqlOptions<P>): any;
export declare function handleResult<R, P>(expr: Expr<R>, result: IResult<ExprValueObjects<R>>, options?: MssqlOptions<P>): any;
export declare function addParams(request: Request, out: DialectOutput, params: any): void;
export declare function addParam(request: Request, out: DialectOutput, paramName: string, paramValue: any): void;
export declare function getDataType(value: any, givenType?: DataTypeInputs): ISqlType | (() => ISqlType);
